﻿[Version = "2.0.0.0"]
section BIMcollab;
///
/// BIMcollab data connector for Microsoft Power BI
/// Copyright © 2018-2020 KUBUS BV"
///
BIMcollab.ConnectorVersion = "2.0"; // friendly version of Version Attribute on first line

BIMcollab.ApiSettings = [
    BasePath = "/data_analysis",
    ApiVersion = "2.0.0",
    PowerBiPrefix = "powerbi-v2-"
];
 
BIMcollab.OAuthSettings = [
    BasePath = "/identity",
    authorization_endpoint= "/connect/authorize",
    token_endpoint = "/connect/token",
    end_session_endpoint= "/connect/endsession",
    scopes = "openid data_analysis offline_access",
    post_logout_redirect_uri = "https://www.bimcollab.com"
];

[DataSource.Kind = "BIMcollab", Publish = "BIMcollab.Publish"]
shared BIMcollab.Contents = let 
    Space.Type = Uri.Type meta [
        Documentation.FieldCaption = Translate.FromKey("Space_FieldCaption"),
        Documentation.FieldDescription = Translate.FromKey("Space_FieldDescription"),
        Documentation.SampleValues = Text.Split(Translate.FromKey("Space_SampleValues"), ";")
    ], 
    ContentsType = type function (space as Space.Type) as table meta [
        Documentation.Name = Text.Combine({ Translate.FromKey("Connector_Name"),  BIMcollab.ConnectorVersion}, " ")
    ]
in
    Value.ReplaceType(BIMcollab.Services, ContentsType);

BIMcollab = [
    TestConnection = (dataSourcePath) => let space = BIMcollab.GetSpaceFromEnv(dataSourcePath) in {"BIMcollab.Contents", space },
    Authentication = [
        OAuth = [
            StartLogin = OAuth.StartLogin,
            FinishLogin = OAuth.FinishLogin,
            Refresh = OAuth.Refresh,
            Logout = OAuth.Logout,
            Label = Translate.FromKey("Authentication_OAuthLabel")
        ]
    ],
    Icons = BIMcollab.Icons,
    Description = Translate.FromKey("BIMcollab")
];

BIMcollab.Publish = [
    Beta = false,
    Category = "Online Services",
    ButtonText = { Translate.FromKey("Connector_ButtonTitle"), Translate.FromKey("Connector_ButtonHelp") },
    LearnMoreUrl = Translate.FromKey("Connector_LearnMoreUrl"),
    SourceImage = [
        Icon32 = { Extension.Contents("BIMcollab32Sheet.png"), Extension.Contents("BIMcollabSheet40.png"), Extension.Contents("BIMcollabSheet48.png"), Extension.Contents("BIMcollabSheet64.png") }
    ],
    SourceTypeImage = BIMcollab.Icons
];


BIMcollab.Icons = [
    Icon16 = { Extension.Contents("BIMcollab16.png"), Extension.Contents("BIMcollab20.png"), Extension.Contents("BIMcollab24.png"), Extension.Contents("BIMcollab32.png") },
    Icon32 = { Extension.Contents("BIMcollab32.png"), Extension.Contents("BIMcollab40.png"), Extension.Contents("BIMcollab48.png"), Extension.Contents("BIMcollab64.png") }
];

BIMcollab.Services = (space as text) =>
let
    ApiUrl = BIMcollab.GetApiBaseUrl(space),
    Path = Text.FormatNamed("{BasePath}/{ApiVersion}/navigationtable", BIMcollab.ApiSettings),
    NavigationList = Json.Document(Web.Contents(ApiUrl, [ RelativePath = Path ])),
    ApiCheck = List.Count(NavigationList) > 0
in
    if not (ApiCheck) then ShowError(Translate.FromKey("Space_Api_CanNotConnect")) 
    else BIMcollab.NavigationTable(ApiUrl, NavigationList);

     
BIMcollab.NavigationTable = (ApiUrl as text, NavigationList as list) as table =>
let 
    Table = BIMcollab.ConvertToTable(NavigationList),
    Data = Table.AddColumns(Table, 
    {         
        { 
            NavigationColumn.Data, (row) => if row[IsLeaf] 
                then BIMcollab.CallApiEndpoint(ApiUrl, row[DataApiPath], row[Fields], row[Key]) 
                else BIMcollab.NavigationTable(ApiUrl, row[Branches]), 
                type table
            }
    }),
    CleanedData = Table.RemoveColumns (Data, { "Branches", "DataApiPath", "Fields" }) 
in 
    NavigationTable.FromTable(CleanedData);

BIMcollab.CallApiEndpoint = (url as text, path as text, optional fields as list, optional ServiceId as text) => let
    Source = Json.Document(Web.Contents(url, [ RelativePath = path, Timeout = #duration(0, 0, 45, 0) ])),
    IsIssueTable = ServiceId = "issues" or ServiceId = "project.issues",
    IsScheduleRecord = Value.Is(Source, type record)  
in
    if IsScheduleRecord then BIMcollab.ConvertScheduleDataToTable(Source)
    else if IsIssueTable then BIMcollab.ConvertIssuesToTable(url, Source, fields)
    else BIMcollab.ConvertToTable(Source, fields);


BIMcollab.ConvertScheduleDataToTable = (data as record) => 
let   
    tempColumnName = Text.NewGuid(),
    rawTable = Table.FromList(data[records], each {_}, { tempColumnName }),
    columnSettings = List.Transform(data[properties], (property) => 
        let 
            columnName = Text.Combine(List.RemoveNulls({ Record.FieldOrDefault(property, "propertySetName"), Record.FieldOrDefault(property, "name") }), "."),
            typeName = IfNull(Record.FieldOrDefault(property, "type"), "none"),
            columnType = Record.FieldOrDefault(Data.ScheduleDataTypes, typeName, type text),
            columnValueFunction = (row) => BIMcollab.ConvertScheduleValue(row, tempColumnName, property[index], columnType)
        in
            { columnName, columnValueFunction, columnType }),

    tableWithColumns = Loop.For(rawTable, columnSettings, Table.AddColumnAutoNumbered),
    tableWithColumnsCleaned = Table.RemoveColumns(tableWithColumns, tempColumnName)
in
   tableWithColumnsCleaned;

BIMcollab.ConvertScheduleValue = (row as record, listColumnName as text, index as number, columnType as type) =>
let
    list = Record.FieldOrDefault(row, listColumnName, {}),
    outOfRange = List.Count(list) <= index,
    value = if outOfRange then null else List.First(List.Skip(list, index)),
    targetType = if Type.IsNullable(columnType) then Type.NonNullable(columnType) else columnType,
    convertedValue = if Value.Is(value, targetType) then value
                     else if (value = null or value = "") and Type.IsNullable(columnType) then null
                     else if Type.Is(targetType, type number) and Value.Is(value, type text) then try Number.FromText(value) otherwise value
                     else if Type.Is(targetType, type text) then Text.ToString(value)
                     else if Type.Is(targetType, type logical) and Value.Is(value, type text) then List.Contains({"1", "true"}, Text.Lower(IfNull(value, "")))
                     else if Type.Is(targetType, type datetimezone) and Value.Is(value, type text) then DateTimeZone.FromText(value)
                     else value
in
    convertedValue;

BIMcollab.ConvertIssuesToTable = (url as text,source as list, fields as list ) => 
    BIMcollab.ConvertToTable(source, fields, {[SourceName = "Id", DisplayName = "Id", OutputType = "long"]}, 
            (table) => let 
                Baseurl = BIMcollab.GetApiBaseUrl(url),
                SourceColumnExists = List.Contains( Table.ColumnNames(table), "Id"),
                GetIssueUrl = (row) => Text.Combine({ Baseurl, "/data_analysis/issue/", Number.ToText(row[Id]) })
            in
                if SourceColumnExists then Table.AddColumn(table, "IssueUrl", GetIssueUrl , type text) else table);


BIMcollab.ConvertToTable = (data as list, optional fields as list, optional tempFields as list, optional postProcessTable as function) => let
    hasData = List.Count(data) > 0, // to return empty table instead of error
    Table = if hasData then Table.FromList(data, Splitter.SplitByNothing(), null, null, ExtraValues.Error) else #table({"_"},{{[Message = "No data available"]}}),    
    
    fields0 = IfNull(fields, List.Transform(Table.GetRecordColumnFieldNames(Table, Table.ColumnNames(Table){0}), (i) => [SourceName = i, DisplayName = i])),
    fields1 = if tempFields <> null then List.Combine({ fields0, tempFields }) else fields0,
    sourceColumns = List.Transform(fields1, (i) => if(Value.Is(i, type record)) then i[SourceName] else i),
    targetColumns = List.Transform(fields1, (i) => if(Value.Is(i, type record)) then i[DisplayName] else i),   
    Data = if hasData then Table.ExpandRecordColumn(Table, Table.ColumnNames(Table){0}, sourceColumns, targetColumns) else #table(targetColumns, {}),
    DataAsTable = Data.ConvertColumns(Data, fields1),
    PostProcessedTable = if postProcessTable <> null then Function.Invoke(postProcessTable, { DataAsTable }) else DataAsTable,
    CleanedTable = Table.RemoveColumns(PostProcessedTable, List.Transform(IfNull(tempFields,{}), (record) => record[DisplayName]))
in
    CleanedTable;

Data.NamedTypes = [
    datetime = type datetimezone,
    datetimeoffset = type datetimezone,
    string = type text,
    int = Int32.Type,
    decimal = Decimal.Type,
    double = Double.Type,
    long = Int64.Type,
    bool = type logical
];

Data.ScheduleDataTypes = [
    boolean = type logical,
    logical = type nullable logical,
    datetime = type nullable datetimezone,
    number= type nullable number,
    string = type text
];

Data.ConvertColumns = (table as table, columns as list) => let
    ShouldTransform = List.Count(columns) > 0 and Value.Is(columns{0}, type record) and Record.FieldOrDefault(columns{0}, "OutputType") <> null,
    Result = if(ShouldTransform) then Table.TransformColumnTypes(table,{{columns{0}[DisplayName], Record.Field(Data.NamedTypes, columns{0}[OutputType])}}) else table
in
    if List.Count(columns) > 1 then Data.ConvertColumns(Result, List.RemoveFirstN(columns, 1)) else Result;

BIMcollab.GetClientId = (space as text) as text
    => BIMcollab.ApiSettings[PowerBiPrefix] & BIMcollab.GetSpaceName(space);

BIMcollab.GetSpaceFromEnv = (env) => if(Text.StartsWith(env, "{")) then Record.Field(Json.Document(env), "space") else env;  

BIMcollab.GetBaseUrl = (space as text) => let
    TrimmedUrl = Text.Lower(Text.Trim(Text.Combine(List.Range(Text.Split(space, "/"), 0, 3), "/"), "/")),
    SchemeRemoved = Text.ReplaceMany(TrimmedUrl, {{"http://", ""}, {"https://", ""}}),
    Hostname = List.First(Text.SplitAny(SchemeRemoved, "/"))
in
   if Text.StartsWith(Hostname, "localhost:") then TrimmedUrl
   else if Hostname <> null then  "https://" & Hostname
   else ShowError(Translate.FromKey("Space_Url_Invalid"));

BIMcollab.GetApiBaseUrl = (space as text) => let
    BaseUrl = BIMcollab.GetBaseUrl(space),
    Query = Uri.Parts(space)[Query],
    ApiUrl = if Record.HasFields(Query, { "apiurl" }) then Query[apiurl] else null
in
    IfNull(ApiUrl, BaseUrl);

BIMcollab.GetOAuthBaseUrl = (space as text) 
    => Uri.Combine(BIMcollab.GetBaseUrl(space), BIMcollab.OAuthSettings[BasePath]);

BIMcollab.GetSpaceName = (space as text) => let
    BaseUrl = BIMcollab.GetBaseUrl(space),
    SpaceName = Text.SplitAny(BaseUrl, "/"){2}
in
    SpaceName;

OAuth.StartLogin = (env, state, display) =>
let     
    space = BIMcollab.GetSpaceFromEnv(env), 
    oauth_base_uri = BIMcollab.GetOAuthBaseUrl(space),
    api_base_uri = BIMcollab.GetApiBaseUrl(space),
    isValid = Text.StartsWith(space, "https://") or Text.Contains(api_base_uri, "localhost:"),
    client_id = BIMcollab.GetClientId(space),
    code_bytes = Encoding.ASCII.GetBytes(Text.NewGuid()),
    code_verifier = Base64.UrlEncode(code_bytes), 
    context = [
        oauth_base_uri = oauth_base_uri,
        issuer = oauth_base_uri,
        client_id = client_id,
        spacename = BIMcollab.GetSpaceName(space),
        audience = BIMcollab.GetBaseUrl(space), 
        redirect_uri = "https://oauth.powerbi.com/views/oauthredirect.html",
        scopes= BIMcollab.OAuthSettings[scopes],
        state = state, 
        nonce = Text.NewGuid(),
        code_verifier = code_verifier
    ],
    AuthorizeUrl = Text.Combine({oauth_base_uri, BIMcollab.OAuthSettings[authorization_endpoint], "?", Uri.BuildQueryString([
        audience = context[audience],
        client_id = context[client_id],
        scope = context[scopes],
        nonce  = context[nonce],
        state = context[state],
        response_type = "code",
        code_challenge = OAuth.GenerateS256CodeChallenge(code_verifier),
        code_challenge_method = "S256",
        redirect_uri = context[redirect_uri]
    ])}),
    Result = if not isValid then error Error.Record(Translate.FromKey("Space_Url_Invalid"), Translate.FromKey("Space_FieldDescription") & " " & Translate.FromKey("Space_Field_Tips"))
        else [
            LoginUri = AuthorizeUrl,
            CallbackUri = context[redirect_uri], 
            WindowWidth = 800,
            WindowHeight = 800,
            Context = context
        ]
in
    Result;
            
OAuth.FinishLogin = (context, callbackUri, state) =>
let
    parts = Uri.Parts(Text.Replace(callbackUri,"#","?"))[Query],
    Data = [
            audience = context[audience],
            client_id = context[client_id],
            redirect_uri = context[redirect_uri],
            grant_type = "authorization_code",
            scopes =  context[scopes],
            code_verifier = context[code_verifier],
            code =  parts[code]
        ],
        result = if Record.HasFields(parts, {"error", "error_description"}) then parts
                 else OAuth.TokenMethod(context[oauth_base_uri], Data),
        hasError = Record.HasFields(result, {"error"})
in
    if not hasError then result else error Error.Record(result[error]);

OAuth.GetTokenInfo = (token as text) =>
let 
    parts = Text.Split(token, ".")
in 
    [ 
        jose_header = Base64.JsonDecode(parts{0}), // [alg="RS256", kid="89225a...", typ="JWT"}
        payload = Base64.JsonDecode(parts{1}) // [aud, iss, exp, sub ... etc ]
    ];
    
GenerateToken = (jose_header, payload) => Base64.JsonEncode(jose_header) & "." & Base64.JsonEncode(payload) & "." & Base64.JsonEncode([]);

OAuth.TokenMethod = (base_uri as text, Data as record) =>
    let         
        Response = Web.Contents(base_uri, [
            RelativePath = BIMcollab.OAuthSettings[token_endpoint],
            Content = Text.ToBinary(Uri.BuildQueryString(Data)),
            Headers = [
                #"Content-type" = "application/x-www-form-urlencoded", 
                #"Accept" = "application/json"
            ],
            ManualStatusHandling = {400}
        ])
    in
        Json.Document(Response); 

OAuth.Logout = (token) => let
    post_logout_redirect_uri = BIMcollab.OAuthSettings[post_logout_redirect_uri],
    Data = [ 
        id_token_hint = token,
        post_logout_redirect_uri = post_logout_redirect_uri
    ],    
    Issuer = () => OAuth.GetTokenInfo(token)[payload][iss]
in 
    if token = null then null
    else Uri.Combine(Issuer(), BIMcollab.OAuthSettings[end_session_endpoint]) & "?" & Uri.BuildQueryString(Data);

OAuth.Refresh = (env, refresh_token) =>
let 
    space = BIMcollab.GetSpaceFromEnv(env),
    Data = [
        client_id = BIMcollab.GetClientId(space),
        grant_type = "refresh_token",
        refresh_token = refresh_token
    ],
    baseUri=BIMcollab.GetOAuthBaseUrl(space), 
    result = OAuth.TokenMethod(baseUri, Data),
    // if refresh token failed the connector would show "Value cannot be null. Parameter name: accessToken." or "The credentials provided are invalid."
    // so we create invalid access_token in return data to force another message: "We couldn't authenticate with the credentials provided. Please try again."
    invalid_result = if (Record.FieldOrDefault(result, "error", "valid") = "invalid_grant") 
        then 
        [
            access_token = GenerateToken([alg="RS256",typ="JWT"],[iss=baseUri,exp=0]), 
            token_type = "Bearer",
            expires_in = -3600, 
            refresh_token = refresh_token,
            refresh_token_expires_in = -3600
        ]
        else null
in 
    if invalid_result <> null then invalid_result else result;

///
/// Crypto and encoding functions
///

OAuth.GenerateS256CodeChallenge = (codeVerifier as text) as text => Base64.UrlEncode(Crypto.Sha256(Encoding.ASCII.GetBytes(codeVerifier)));

Encoding.ASCII.GetBytes = (s as text) as binary => Text.ToBinary(s, TextEncoding.Ascii);

Crypto.Sha256 = (bytes as binary) as binary => Crypto.CreateHash(CryptoAlgorithm.SHA256, bytes);

Base64.Encode = (bytes as binary) as text => Binary.ToText(bytes, 0);

Base64.UrlEncode = (bytes as binary) as text => let
    Base64Encoded = Base64.Encode(bytes),
    UrlEncoded = Text.ReplaceMany(Base64Encoded, {{ "+", "-" } , { "/",  "_" }, { "=", "" }})
in
    UrlEncoded;

// Decode Base64.(Url)Encoded strings
Base64.Decode = (text as text) as binary => let
    AddPadding = { "", "*", "==" , "=" }{ Number.Mod(Text.Length(text), 4) }, // *: invalid base64(url)
    UndoUrlEncode = Text.ReplaceMany(text, {{ "-", "+" }, { "_",  "/" }})
in
    Binary.From(UndoUrlEncode & AddPadding, 0);     

Base64.JsonDecode = (base64 as text) as record
    => Json.Document(Text.FromBinary(Base64.Decode(base64), TextEncoding.Utf8));

Base64.JsonEncode = (data as record) as text
    => Base64.UrlEncode(Json.FromValue(data));

//
// Common functions
//

Translate.FromKey = (key as text) as text => let
    Value = Extension.LoadString(key)
in
    IfNull(Value, "[" & key & "]");
    
IfNull = (value as any, alt_value as any) as any
    => if value <> null then value else alt_value;

NavigationColumn.Key = "Key";
NavigationColumn.Name = "Name";
NavigationColumn.Data = "Data";
NavigationColumn.ItemKind = "ItemKind";
NavigationColumn.Delay = "ItemName";
NavigationColumn.IsLeaf = "IsLeaf";

NavigationTable.FromTable = (table as table) =>
    let
        tableType = Value.Type(table),
        newTableType = Type.AddTableKey(tableType, {NavigationColumn.Key}, true) meta
        [
            NavigationTable.NameColumn = NavigationColumn.Name,
            NavigationTable.DataColumn = NavigationColumn.Data,
            NavigationTable.ItemKindColumn = NavigationColumn.ItemKind,
            Preview.DelayColumn = NavigationColumn.Delay,
            NavigationTable.IsLeafColumn = NavigationColumn.IsLeaf
        ]
    in
        Value.ReplaceType(table, newTableType);

Table.AddColumns = (table as table, argumentList as list) => Loop.For(table, argumentList, Table.AddColumn);

Table.AddColumnAutoNumbered = (table as table, newColumnName as text, columnGenerator as function, optional columnType) => 
    Table.AddColumn(table, AutoNumber(Table.ColumnNames(table), newColumnName), columnGenerator, columnType);

AutoNumber=(allColumns as list, newColumnName as text) =>
    let
        appendNumber = (value as text, number) => Text.Combine({value, " (", Text.ToString(number), ")"}),
        number = Int32.TryParse(Text.TrimEnd(List.Last(Text.Split(newColumnName," (")),")"))[Value],
        appendix = appendNumber("", number),
        strippedColumnName = if Text.EndsWith(newColumnName, appendix) then Text.RemoveRange(newColumnName, Text.Length(newColumnName) - Text.Length(appendix), Text.Length(appendix)) else newColumnName,
        newColumnName2 = appendNumber(strippedColumnName, List.Max({number + 1, 2})),
        result = if List.Contains(allColumns, newColumnName) then AutoNumber(allColumns, newColumnName2) else newColumnName
    in
        result;


Text.ToString = (value as any) => Text.Format("#{0}", {value});

Text.FormatNamed = (text as text, args as record) => let
    GetValue = (record, key) => try Text.From(Record.Field(record, key)) otherwise "[" & key & "]",
    argumentList = List.Transform(Record.FieldNames(args), (key) => { "{" & key & "}",GetValue(args, key)})
in
    Loop.For(text, argumentList, Text.Replace); 

Int32.TryParse = (value as text) => 
    let
        result = try Int32.From(value)
    in
        if result[HasError] then [Success=false, Value=0] else [Success=true, Value=result[Value]];

Table.GetRecordColumnFieldNames = (table as table, recordColumnName as text) as list => let
    sample = Table.FirstValue(Table.SelectColumns(Table.SelectRows(table, (r) => Record.Field(r, recordColumnName) <> null), recordColumnName))
in
    Record.FieldNames(sample);

Text.ReplaceMany = (text as text, argumentList as list) => Loop.For(text, argumentList, Text.Replace);

// Loop a function for a property for each (list of) argument(s) in a list
Loop.For = (initialValue as any, argumentList as list, func as function) => let
    args = () as list => 
        if not Value.Is(argumentList{0}, type list) then { initialValue, argumentList{0} } // only 2 arguments
        else List.InsertRange(argumentList{0}, 0, { initialValue }), // add initialValue as first argument to list
    result = (args) => Function.Invoke(func, args)
in
    if List.Count(argumentList) > 0 
    then Loop.For(result(args()), List.RemoveFirstN(argumentList, 1), func)
    else initialValue;

ShowError = (errorMessage as text) => error Error.Record("Error", errorMessage);
